# ------------------------------------------------------------------------------
#  Frame: .editor.notebook.instance
#
#  Methods:
#     - InstanceMenu:Create
#     - InstanceMenu:Init
#     - InstanceMenu:CreateOptions
#     - InstanceMenu:CreateTrans
#     - InstanceMenu:SetTransforms
#     - InstanceMenu:SetSLIDETransforms
#     - InstanceMenu:SetParameterValues
#     - InstanceMenu:Adjust
#
#     - InstanceMenu:SelectTransform
#     - InstanceMenu:DropTransformCmd
#     - InstanceMenu:AddTransformCmd
#     - InstanceMenu:RemoveTransformCmd
#
#     - InstanceMenu:UndoCmd
#     - InstanceMenu:RedoCmd
#     - InstanceMenu:ApplyCmd
#     - InstanceMenu:ApplyAlert
#
#     - InstanceMenu:UndoList
#     - InstanceMenu:RedoList
#
#
#  Variables:
#     - scaleLabel translateLabel rotateLabel noTransLabel  = labels for listbox items
#     - labelStart = number of spaces before label
#
#     - instanceTransFrame
#     - instanceOptionsFrame
#     - instLB                = ListBox frame
#     - highlightedTrans
#     - transName             = for enumerating transform frames (0->)
#     - [frame]x, [frame]y, [frame]z, [frame]t, 
#
#     - instanceUndoList
#     - instanceRedoList
#
# ------------------------------------------------------------------------------

proc InstanceMenu:Create { instanceFrame } {
    global instanceTransFrame instanceOptionsFrame

    #########################################
    # Init
    InstanceMenu:Init

    #########################################
    # Add Widgets to pages of notebook

    set instanceTransFrame [Notebook:frame $instanceFrame Transforms]
    InstanceMenu:CreateTrans $instanceTransFrame

    InstanceMenu:SetTransforms {{none}} 1
#    InstanceMenu:SetTransforms { {rotate 1 20 30 4.0} {scale 3.0 4.0 5 60} } 1

    set instanceOptionsFrame [Notebook:frame $instanceFrame Options]
    InstanceMenu:CreateOptions $instanceOptionsFrame
}

proc InstanceMenu:Init { } {
    global scaleLabel translateLabel rotateLabel noTransLabel labelStart 
    global instanceUndoList instanceRedoList

    set labelStart     2
    set scaleLabel     "  scale                       "
    set translateLabel "  translate                   "
    set rotateLabel    "  rotate                      "
    set noTransLabel   "  no transforms"

    set instanceUndoList {}
    set instanceRedoList {}


    global shadingValue shadingList shadingHelp
    global lodValue lodList lodHelp
    global ribbeginValue ribbeginHelp
    global ribendValue ribendHelp
    
    set shadingValue inherit
    set shadingList {inherit wire flat gouraud phong}
    set shadingHelp "This option defines how the polygons of the object are shaded:
   inherit = render in the same way as my parent
   wire = render a wireframe
   flat = render with flat shading 
   gouraud = render with Gouraud shading
   phong = render with Phong shading"

    set lodValue full
    set lodList {off bound edges full}
    set lodHelp "This option defines the quality of the rendering for this subtree:
   off = do not render
   bound = only render the bounding box
   edges = render in wireframe mode
   full = fully render"

    set ribbeginValue ""
    set ribbeginHelp "Renderman specific ASCII text that is written just 
before the primitive when outputting a RIB file"

    set ribendValue ""
    set ribendHelp "Renderman specific ASCII text that is written just 
after the primitive when outputting a RIB file"
}

proc InstanceMenu:CreateOptions { base } {

    set options {surface shading lod}

    set y 20
    foreach option $options {

	CreateOptionParameterFrame $base $option $y
	incr y 60
    }
    incr y 60
    CreateRibEntryFrame $base $y
}

proc InstanceMenu:CreateTrans { base } {
    global instLB highlightedTrans transName paleYellow buttonFont labelFont
    
    set y 20
    set x 50
     
    Label $base.label -borderwidth 1 -font $labelFont -text "transforms:" \
	    -helptext "list of transforms applied top down to the instance"
    Label $base.label2 -borderwidth 1 -font $labelFont \
	    -text "x          y          z         angle"

    set sw [ScrolledWindow $base.transforms -relief sunken]
    set instLB [ListBox $sw.lb -deltay 34 -padx 5 -winpadx 100 \
	    -background white -selectbackground $paleYellow -selectthickness 5 \
	    -selectforeground black\
	    -highlightthickness 0 \
	    -dragevent 1 -dropenabled 1 -dragenabled 1 -dropovermode p -redraw 1\
	    -dropcmd {InstanceMenu:DropTransformCmd} \
	    -droptypes { LISTBOX_ITEM {copy {} move {} link {}} }]
    $sw setwidget $instLB
    $instLB bindText <ButtonPress-1> "InstanceMenu:SelectTransform"

    set highlightedTrans none
    set transName 0

    place $base.label      -x 10  -y $y -anchor nw -bordermode ignore 
    place $base.label2     -x 135 -y $y -anchor nw -bordermode ignore 
    set y [expr $y + 30]
    place $base.transforms -x 10 -y $y -width 375 -height 170 -anchor nw -bordermode ignore 

    # ADD REMOVE Buttons
    Button $base.rem -text "Remove" -command "InstanceMenu:RemoveTransformCmd" \
	    -font $buttonFont
    menubutton $base.add -text "Add" -menu $base.add.menu -relief raised \
	    -font $buttonFont
    set m [menu $base.add.menu -tearoff 0]
    $m add command -label "rotate"    -command "InstanceMenu:AddTransformCmd rotate"
    $m add command -label "scale"     -command "InstanceMenu:AddTransformCmd scale"
    $m add command -label "translate" -command "InstanceMenu:AddTransformCmd translate"

    place $base.add -x 10  -y 225 -width 185 -anchor nw -bordermode ignore 
    place $base.rem -x 200 -y 224 -width 188 -anchor nw -bordermode ignore 

    # APPLY UNDO Buttons
    Button $base.apply -text "Apply" -command "InstanceMenu:ApplyCmd" -font $buttonFont \
	    -state disabled -helptext "Apply the current settings"
    Button $base.undo  -text "Undo"  -command "InstanceMenu:UndoCmd"  -font $buttonFont \
	    -state disabled -helptext "Undo last Apply"
    Button $base.redo  -text "Redo"  -command "InstanceMenu:RedoCmd"  -font $buttonFont \
	    -state disabled -helptext "Redo last Undo"

    place $base.apply -x 10  -y 290 -width 100 -anchor nw -bordermode ignore 
    place $base.undo  -x 150 -y 290 -width 100 -anchor nw -bordermode ignore 
    place $base.redo  -x 288 -y 290 -width 100 -anchor nw -bordermode ignore 
}

proc InstanceMenu:CreateTransformEntries { name type x y z theta} {
    global instLB [subst $name]x [subst $name]y [subst $name]z [subst $name]t
    global scaleLabel translateLabel rotateLabel noTransLabel 
    global textHighlightColor labelFont entryFont

    frame $instLB.$name -width 252 -height 30 -background white 

    Entry $instLB.$name.x -textvariable [subst $name]x -font $entryFont \
	   -selectbackground $textHighlightColor
    Entry $instLB.$name.y -textvariable [subst $name]y -font $entryFont \
	   -selectbackground $textHighlightColor
    Entry $instLB.$name.z -textvariable [subst $name]z -font $entryFont \
	   -selectbackground $textHighlightColor

    place $instLB.$name.x -x 0   -y 4 -width 50 -height 22 -anchor nw 
    place $instLB.$name.y -x 60  -y 4 -width 48 -height 22 -anchor nw 
    place $instLB.$name.z -x 120 -y 4 -width 48 -height 22 -anchor nw 

    bind $instLB.$name   <ButtonPress-1> "InstanceMenu:SelectTransform $name"

    bind $instLB.$name.x <ButtonPress-1> "InstanceMenu:SelectTransform $name"
    bind $instLB.$name.y <ButtonPress-1> "InstanceMenu:SelectTransform $name"
    bind $instLB.$name.z <ButtonPress-1> "InstanceMenu:SelectTransform $name"
    bind $instLB.$name.x <KeyPress>      "InstanceMenu:ApplyAlert"
    bind $instLB.$name.y <KeyPress>      "InstanceMenu:ApplyAlert"
    bind $instLB.$name.z <KeyPress>      "InstanceMenu:ApplyAlert"

    if { $type == "rotate" } {
	Entry $instLB.$name.t -textvariable [subst $name]t -font $entryFont \
	   -selectbackground $textHighlightColor

	place $instLB.$name.t -x 190 -y 4 -width 48 -height 22 -anchor nw 
	bind $instLB.$name.t <ButtonPress-1> "InstanceMenu:SelectTransform $name"
	bind $instLB.$name.t <KeyPress>      "InstanceMenu:ApplyAlert"
    }

    if { $type == "scale" } {
	set text $scaleLabel
    } elseif { $type == "translate" } { 
	set text $translateLabel
    } elseif { $type == "rotate" } { 
	set text $rotateLabel
    } else {
	set text $noTransLabel
	$instLB insert end t0 -font $labelFont -text $text
	return;
    }
    
    $instLB insert end $name -font $labelFont -text $text -window $instLB.$name
    
    global [subst $name]x [subst $name]y [subst $name]z [subst $name]t
    set [subst $name]x $x
    set [subst $name]y $y
    set [subst $name]z $z
    set [subst $name]t $theta
}

proc InstanceMenu:SetTransforms { transforms resetUndoList } {
    global instLB transName instanceUndoList instanceRedoList instanceTransFrame

    if { $resetUndoList == 1 } {
	# reset undo list
	set instanceUndoList {}
	set instanceRedoList {}
	InstanceMenu:UndoList push $transforms	

	$instanceTransFrame.undo  configure -state disabled
	$instanceTransFrame.apply configure -state disabled
	$instanceTransFrame.redo  configure -state disabled
    }
    
    # Delete all items    
    set oldTransforms [$instLB items]
    
    foreach transFrame $oldTransforms {
	destroy $instLB.$transFrame
    }
    $instLB delete $oldTransforms
    set transName 0

    ####
    
    set i 0
    foreach transform $transforms {
	
	set name  t[subst $i]
	set type  [lindex $transform 0]
	set x     [lindex $transform 1]
	set y     [lindex $transform 2]
	set z     [lindex $transform 3]
	set theta [lindex $transform 4]

	InstanceMenu:CreateTransformEntries $name $type $x $y $z $theta
	
	#set itemExists [$instLB exists $name]
	#$instLB itemconfigure $name -text $text
	
	incr i 1
    }
}

proc InstanceMenu:SetParameterValues { paramValues } {
    global surfaceValue shadingValue lodValue ribbeginValue ribendValue

    foreach pair $paramValues {

	set param [lindex $pair 0]
	set value [lindex $pair 1]

	if { $value == "(null)" } {
	    set value ""
	}

	#puts "$param = $value"	
	set [subst $param]Value $value
    }
}

proc InstanceMenu:SelectTransform { selectedTransform } {
    global instLB highlightedTrans noTransLabel paleYellow

    set text [$instLB itemcget $selectedTransform -text]

    if { $text == $noTransLabel } {

	return
    }

    # unhighlight old selected transform
    if { ($highlightedTrans != "none") && [$instLB exists $highlightedTrans] } {
	$instLB.$highlightedTrans configure -background white
    }
    
    # highlight selected transform
    $instLB selection set $selectedTransform
    $instLB.$selectedTransform configure -background $paleYellow
    set highlightedTrans $selectedTransform
}

proc InstanceMenu:DropTransformCmd { args } {
    global instLB highlightedTrans

    set pathName      [lindex $args 0]
    set selectedTrans [lindex $args 5]
    set dropPosition  [lindex [lindex $args 2] 1]
    set oldPosition   [$pathName index $selectedTrans]

    if { $dropPosition > $oldPosition } {
	set dropPosition [expr $dropPosition -1]
    }

    $pathName move $selectedTrans $dropPosition

    #$instLB selection clear $selectedTrans

    InstanceMenu:ApplyAlert
}

proc InstanceMenu:UndoList { cmd transforms } {
    global instanceUndoList instanceTransFrame

    if { $cmd == "push" } {

	lappend instanceUndoList $transforms

	if { 1 < [llength $instanceUndoList] } {
	    
	    $instanceTransFrame.undo configure -state normal
	}

    } elseif { $cmd == "pop" } {

	set lastIndex [expr [llength $instanceUndoList] - 1]
	set transforms [lindex $instanceUndoList $lastIndex]
	set instanceUndoList [lreplace $instanceUndoList $lastIndex $lastIndex]
	InstanceMenu:RedoList push $transforms

	set transforms [lindex $instanceUndoList [expr $lastIndex-1]]
	InstanceMenu:SetTransforms $transforms 0

	#	if { $instanceUndoList == "{{none}}" } {}
	if { 1 == [llength $instanceUndoList] } {
	    
	    $instanceTransFrame.undo configure -state disabled
	}
    }
    puts "undo: $instanceUndoList"
    return $transforms
}

proc InstanceMenu:RedoList { cmd transforms } {
    global instanceRedoList instanceTransFrame

    if { $cmd == "push" } {
	lappend instanceRedoList $transforms
	$instanceTransFrame.redo configure -state normal

    } elseif { $cmd == "pop" } {

	set lastIndex [expr [llength $instanceRedoList] - 1]
	set transforms [lindex $instanceRedoList $lastIndex]
	set instanceRedoList [lreplace $instanceRedoList $lastIndex $lastIndex]
	InstanceMenu:UndoList push $transforms

	InstanceMenu:SetTransforms $transforms 0

	if { $instanceRedoList == "" } {
	    
	    $instanceTransFrame.redo configure -state disabled
	}
    }
    puts "redo: $instanceRedoList"
    return $transforms
}

proc InstanceMenu:UndoCmd { } {
    global instanceTransFrame

    if { "disabled" == [$instanceTransFrame.apply cget -state] } {

	puts "pop transform"

	# pop transform
	set transforms [InstanceMenu:UndoList pop 0]
    } else {

	global instanceUndoList

	# throw away latest unapplied changes
	set lastIndex [expr [llength $instanceUndoList] - 1]
	set transforms [lindex $instanceUndoList $lastIndex]
	InstanceMenu:SetTransforms $transforms 0

	puts "throw away current and go back to tansform: $transforms"
    }
    InstanceMenu:SetSLIDETransforms $transforms
}

proc InstanceMenu:RedoCmd { } {

    set transforms [InstanceMenu:RedoList pop 0]
    InstanceMenu:SetSLIDETransforms $transforms
}

proc InstanceMenu:ApplyAlert { } { 
    global instanceTransFrame

    $instanceTransFrame.undo configure -state normal

    $instanceTransFrame.apply configure -state normal
    $instanceTransFrame.apply configure -foreground red
    $instanceTransFrame.apply configure -activeforeground red
}

proc InstanceMenu:ApplyCmd { } {
    global instLB editNodeName noTransLabel editStandAlone instanceTransFrame

    # unalert
    $instanceTransFrame.apply configure -state disabled
    $instanceTransFrame.apply configure -foreground black
    $instanceTransFrame.apply configure -activeforeground black

    # error check, set SLIDE, &  put on undo list
    set transforms [$instLB items]
    set argc [llength $transforms]
    set firstLabel [$instLB itemcget [lindex [split $transforms] 0] -text]

    # no transforms
    if { $firstLabel == $noTransLabel } {

	set undoTransforms {{none}}

    } else {

	set undoTransforms {}
	set i 0
	foreach trans $transforms {
	    global [subst $trans]x [subst $trans]y [subst $trans]z [subst $trans]t
	    global labelStart
	    
	    set type [$instLB itemcget $trans -text]
	    # This is to get rid of the white space
	    set type [lindex [split $type] $labelStart]
	    set x [expr $[subst $trans]x]
	    set y [expr $[subst $trans]y]
	    set z [expr $[subst $trans]z]
	    set t [expr $[subst $trans]t]
	    
	    # Error check
	    if { $type == "scale" } {
		
		if { ($x == 0) && ($y == 0) && ($z == 0) } {
		    set x 0.1
		    set [subst $trans]x $x
		}
	    }

	    lappend undoTransforms "$type $x $y $z $t"
	}
    }

    InstanceMenu:SetSLIDETransforms $undoTransforms
    InstanceMenu:UndoList push $undoTransforms
}

proc InstanceMenu:SetSLIDETransforms { transforms } {
    global editNodeName editStandAlone

    # update SLIDE
    set argc [llength $transforms]

    # no transforms
    if { "none" == [lindex [lindex $transforms 0] 0] } {

	set cmd "slideSetParameters $editNodeName instance 1 transform 0"

    } else {

	set i 0
	foreach trans $transforms {
	    
	    set type [lindex $trans 0]
	    set x    [lindex $trans 1]
	    set y    [lindex $trans 2]
	    set z    [lindex $trans 3]
	    set a    [lindex $trans 4]
	    
	    set arg[subst $i] $type
	    incr i 1
	    set arg[subst $i] $x
	    incr i 1
	    set arg[subst $i] $y
	    incr i 1
	    set arg[subst $i] $z
	    incr i 1
	    set arg[subst $i] $a
	    incr i 1
	}
	
	set cmd "slideSetParameters $editNodeName instance 1 transform $argc"
	for {set i 0} { $i < [expr $argc*5]} {incr i 1} {
	    
	    set cmd [concat $cmd "\$arg[subst $i]" ]
	}
    }

    if { $editStandAlone == 0 } {
	eval $cmd
    }
}

proc InstanceMenu:AddTransformCmd { type } {
    global instLB transName noTransLabel
 
    # check if need to delete "no transform" item
    if { 1 == [llength [$instLB items]] } {

	set trans [lindex [$instLB items] 0]
	set label [$instLB itemcget $trans -text]
	if { $label == $noTransLabel } {
	    
	    destroy $instLB.$trans
	    $instLB delete $trans
	}
    }

    if { $type == "scale" } {
	InstanceMenu:CreateTransformEntries a[subst $transName] $type 1 1 1 0
    } elseif { $type == "translate" } {
	InstanceMenu:CreateTransformEntries a[subst $transName] $type 0 0 0 0
    } elseif { $type == "rotate" } {
	InstanceMenu:CreateTransformEntries a[subst $transName] $type 1 0 0 0
    }

    InstanceMenu:SelectTransform a[subst $transName]
    InstanceMenu:ApplyAlert

    incr transName 1
}

proc InstanceMenu:RemoveTransformCmd { } {
    global instLB highlightedTrans

    # delete all item and window
    destroy $instLB.$highlightedTrans
    $instLB delete $highlightedTrans

    set highlightedTrans "none"

    if { "" == [$instLB items] } {

	InstanceMenu:SetTransforms {{none}} 0
    }

    InstanceMenu:ApplyAlert
}
