source graphglobals.tcl
source graphinit.tcl

#########################################################################################
# ------------------------------------------------------------------------------
#  Frame: .graph
#
#  Methods:
#     - Graph:Create
#     - Graph:Init
#     - Graph:CreateScrolledCanvas
#     - Graph:ScrollY
#     - Graph:ScrollX
#
#     - Graph:DrawHighlight
#     - Graph:DrawInstance
#     - Graph:DrawReference
#     - Graph:RedrawHeadNodeInstances
#     - Graph:RedrawTailNodeInstances
#
#     - Graph:CreateReference
#     - Graph:CreateInstance
#     - Graph:CreateInstanceLine
#     - Graph:CreateInstanceCircle
#     - Graph:RegisterInstance
#     - Graph:CanvasMiddleRelease
#     - Graph:CanvasMiddleDrag
#     - Graph:CanvasMiddleSelect
#
#     - Graph:LeftDragSelected
#     - Graph:LeftSelect
#     - Graph:LeftRelease
#     - Graph:LeftDown
#     - Graph:Motion
#
#     - Graph:CreateNode
#     - Graph:PaletteCreateNode
#     - Graph:SLIDEObjectSelected = an object was selected in SLIDE
#     - Graph:ObjectSelected      
#     - Graph:ObjectUnselected
#
#     - Graph:SelectedPathAdd
#     - Graph:SelectedPathClear
#
#     - Graph:EntryShow    = replace name text with entry box
#     - Graph:EntryHide    = hide entry box and modify name text
#     - Graph:EntryProcess = check if entry name is ok to use in SLIDE
#
#     - Graph:Delete
#     - Graph:DeleteInstance
#     - Graph:DeleteNode
#     - Graph:Clear
#     - Graph:Destory
#
#  Variables:
#     - canvas             = array for mouse-selected objectId, x, y, xscroll, yscroll
#     
#     - paletteSelection   = "none" or type selected from palette
#
#     - selectedObject     = nodeId of selectedObject
#     - selectedPath       = path of selectedObject
#
#     - objectSelected     = bool: 1 - an object is selected with left release
#     - keepSelected       = bool: 1 - keep the selected object selected
#     - dragged            = bool: 1 - left mouse was dragged (used for selection) 
#
#     - instanceName       = current instance being drawn (none otherwise)
#     - tailNodeId         = tail node of instanceName
#     - xTail, yTail       = x and y positoin of tailNodeId
#
#     - graphEntryId       = entry id
#     - graphEntryTextId   = id of text that was replaced with the entry box
#     - graphEntryHasFocus = entry is up and has focus
#     - userName(name)     = array of names that are in use
#
#     - nodeSLIDEId(id)            = nodeName, instName
#     - instances(tailId, headId)  = {instanceNames between tail and head}
#     - nodeInstanceTails(nodeID)  = {instanceNames whose head points to the node}
#     - nodeInstanceHeads(nodeID)  = {instanceNames whose tail leaves the node}
#
#     - instanceTail(instanceId)   = tailId
#                                      |
#                                      V
#     - instanceHead(instanceId)   = headId
#
# ------------------------------------------------------------------------------

proc Graph:SetSLIDEGraphPositions { } {
    global nodeSLIDEId 

    set objectIds [.graph.canvas find withtag "movable"]

    foreach id $objectIds {

	set name $nodeSLIDEId($id)
	set type   [lindex [.graph.canvas gettags $id] 0]
	set thisXY [.graph.canvas coords $id]
	set x [expr int([lindex $thisXY 0])]
	set y [expr int([lindex $thisXY 1])]

	#puts "slideSetParameters $name $type 0 graphposition $x $y"

	if { ($type == "group") || ($type == "instance") || \
		($type == "sphere") || ($type == "cone") ||  ($type == "cylinder") || \
		($type == "torus") ||  ($type == "cyclide") } {  

	    slideSetParameters $name $type 0 graphposition $x $y
	}
    }
}

proc Graph:SelectedPathAdd { instName } {
    global selectedPath viewerSelected

    set viewerSelected 1

    lappend selectedPath $instName
    puts "selectedPath: $selectedPath"

    .graph.canvas itemconfigure [subst $instName]Line -fill red 
}

proc Graph:SelectedPathClear { clearObject } {
    global selectedPath arrowColor viewerSelected

    foreach instName $selectedPath {
	
	.graph.canvas itemconfigure [subst $instName]Line -fill $arrowColor
    }
    set selectedPath {}
    #puts "selectedPath: $selectedPath"
    
    # When viewer unselects
    if { $clearObject == 1 } {
	Graph:ObjectUnselected 0
	set viewerSelected 0
    }
}

#########################################################################################
# Clear Graph

proc Graph:Destroy { } {

    destroy .graph
}

proc Graph:Clear { } {

    set nodeIds [.graph.canvas find withtag movable]

    foreach id $nodeIds {

	Graph:DeleteNode $id
    }

    Graph:InitCanvasSetting
}

proc Graph:DeleteNode { nodeId } {
    global canvas nodeSLIDEId selectedObject objectSelected
    global nodeInstanceTails nodeInstanceHeads userName

    # Delete all instances that are connected to this node
    set mynodeInstanceTails $nodeInstanceTails($nodeId)
    set mynodeInstanceHeads $nodeInstanceHeads($nodeId)
    
    foreach inst $mynodeInstanceTails {
	if { $inst != "empty" } {

	    Graph:DeleteInstance $inst
	}
    }
    foreach inst $mynodeInstanceHeads {
	if { $inst != "empty" } {
	    
	    Graph:DeleteInstance $inst
	}
    }
    
    # Delete this node from all node arrays
    set empty {empty}
    set nodeInstanceTails($nodeId) $empty
    set nodeInstanceHeads($nodeId) $empty

    set nodeName [lindex [.graph.canvas gettags $nodeId] 1]
    set userName($nodeName) 0
    
    # Delete the node
    .graph.canvas delete $nodeId
    
    # Delete the label
    set textNode text[subst $nodeSLIDEId($nodeId)]
    .graph.canvas delete $textNode
    
    if { $selectedObject == $nodeId } {
	
	Graph:ObjectUnselected 1
    }
}


# Clear Graph
#########################################################################################

#########################################################################################
# Delete Node

proc Graph:Delete { mouseX mouseY } {
    global canvas nodeSLIDEId objectSelected graphStandAlone graphEntryHasFocus
    
    if { ($graphEntryHasFocus == 0) && ($objectSelected == 1) } {
	
	set nodeId $canvas(obj)
	set isThisAnInstance [string first instance $[lindex [.graph.canvas gettags $nodeId] 0]]
	set name $nodeSLIDEId($nodeId)
	set type [lindex [.graph.canvas gettags $nodeId] 0]
	
	if { ($name == "WORLD") } {
	    
	    puts "Can't delete $name"
	    return
	}
	
	if { $isThisAnInstance != -1 } {
	    
	    set nodeId $nodeSLIDEId($nodeId)  
	    Graph:DeleteInstance $nodeId
	} else {
	    
	    Graph:DeleteNode $nodeId
	}
	
	if { ($graphStandAlone == 0) } {

	    #puts "---------------------- slideDelete $name $type"

	    #slideDelete $name $type
             slide delete $type $name 
	}
    }
}

proc Graph:DeleteInstance { instName } {
    global instances nodeInstanceTails nodeInstanceHeads instanceHead instanceTail
    global nodeSLIDEId selectedObject objectSelected userName

    puts "deleting $instName"

    set tailNodeId $instanceTail($instName)
    set headNodeId $instanceHead($instName)

    # Delete this instance from node arrays    
    set nodeInstanceTails($tailNodeId) \
	    [ldelete $nodeInstanceTails($tailNodeId) $instName]
    set nodeInstanceHeads($headNodeId) \
	    [ldelete $nodeInstanceHeads($headNodeId) $instName]

    # Delete this instance from instance arrays
    set userName($instName) 0
    set instanceTail($instName) empty
    set instanceHead($instName) empty
    set instances($tailNodeId,$headNodeId) \
	    [ldelete $instances($tailNodeId,$headNodeId) $instName]

    # Redraw instances
    Graph:RedrawHeadNodeInstances $headNodeId

    # Delete this instance, instanceCircle, and label
    .graph.canvas delete [subst $instName]Line
    .graph.canvas delete $instName
    .graph.canvas delete text[subst $instName]

    if { ($objectSelected == 1) && ($instName == $nodeSLIDEId($selectedObject)) } {

	Graph:ObjectUnselected 1
    }
}

# Delete Node
#########################################################################################

#########################################################################################
# Adding New Nodes

proc Graph:PaletteCreateNode { type } {
    global canvas paletteSelection keepSelected

    # Delete the node that was created but not placed on the canvas
    if { $paletteSelection != "none" } {

	set nodeId $canvas(obj)
	Graph:DeleteNode $nodeId
    }

    set paletteSelection $type
    set x 0
    set y -500
    set nodeName $type[subst [Graph:GetNodeId $type]]

    set nodeId [Graph:CreateNode $type $nodeName $x $y]
 
    # Make node follow the mouse
    set canvas(obj) $nodeId
    set canvas(x) $x
    set canvas(y) $y

    # so the wrong node won't get deleted
    #Graph:ObjectSelected $nodeId

    set keepSelected 1
}

proc Graph:CreateNode { type nodeName x y } {
    global imageWidth imageHeight textHeight textHighlightColor
    global canvas paletteSelection nodeInstanceTails nodeInstanceHeads
    global nodeSLIDEId instancableTypes userName

    set tags $type
    lappend tags $nodeName
    lappend tags movable

    if { -1 != [lsearch $instancableTypes $type] } {

	lappend tags instancable
    }

    # image
    set nodeId [.graph.canvas create image $x $y -image $type -anchor nw -tag $tags]

    # text
    set tags text[subst $nodeName]
    set textId [.graph.canvas create text \
	    [expr $x + $imageWidth/2.0] [expr $y + $textHeight] \
	    -text $nodeName -justify center -tag $tags]
    .graph.canvas bind $textId <Button-1>   "Graph:EntryShow $textId $nodeName"

    set empty {empty}
    lappend nodeInstanceTails($nodeId) $empty
    lappend nodeInstanceHeads($nodeId) $empty
    set nodeSLIDEId($nodeId) $nodeName
    set userName($nodeName) 1

    return $nodeId
}

proc Graph:Motion { mouseX mouseY } {
    global canvas paletteSelection imageWidth imageHeight
    global canvasTotalWidth canvasTotalHeight canvasWidth canvasHeight
    global .graph.xscroll .graph.yscroll
    global nodeSLIDEId

    #global selectedObject
    #puts "so: $selectedObject"

    if { $paletteSelection != "none" } {

	set x [expr $mouseX + $canvas(xscroll)]
	set y [expr $mouseY + $canvas(yscroll)]
	set dx [expr $x - $canvas(x)]
	set dy [expr $y - $canvas(y)]

	set nodeId $canvas(obj)
	set textName $nodeSLIDEId($nodeId)
	set textId [.graph.canvas find withtag text[subst $textName]]

	.graph.canvas move $nodeId $dx $dy
	.graph.canvas move $textId $dx $dy
	set canvas(x) $x
	set canvas(y) $y
    }
}

proc Graph:LeftDown { mouseX mouseY } {
    global paletteSelection selectedObject objectSelected keepSelected
    global canvas graphStandAlone entryHideCalled graphEntryHasFocus

    #puts "leftdown"
    #set ids [.graph.canvas find all]
    #puts $ids

    # placing new node
    if {$paletteSelection != "none"} {
	
	# add this node to SLIDE
	set nodeId $canvas(obj)
	set nodeName [lindex [.graph.canvas gettags $nodeId] 1]

	if { $graphStandAlone == 0 } {

	    #puts "----------------------  slideAddNode $nodeName $paletteSelection"

	    #slideAddNode $nodeName $paletteSelection
            slide create $paletteSelection $nodeName 
	}

	# reset paletteSelection
	set paletteSelection none
    } else {
	Graph:SelectedPathClear 0
    }


    if { ($objectSelected == 1) && ($keepSelected == 0) } {

	set keepSelected 0
	if { ($entryHideCalled == 0) && ($graphEntryHasFocus == 1) } {

	    set keepSelected [Graph:EntryProcess]
	}

	if { $keepSelected == 0 } {
	    Graph:ObjectUnselected 1
	}
    } 

    set keepSelected 0
    set entryHideCalled 0
}

proc Graph:EntryProcess { } {
    global editorNameFrame entryHideCalled entryHideError

    set entryHideCalled 1
    set entryHideError  0

    set results [Global:CheckSLIDEName]
    set type    [lindex $results 0]
    set i       [lindex $results 1]
    
    if { $type == "bad" } {
	
	# highlight incorrect name
	.graph.canvas.entryName selection from $i
	.graph.canvas.entryName selection adjust end

	return 1
    }
    
    ########################################

    Graph:EntryHide 

    return 0
}

proc Graph:EntryHide { } {
    global graphEntryHasFocus graphEntryId graphEntryTextId userName editNodeUserName 

    focus .graph
    set graphEntryHasFocus 0	
	
    set textXY [.graph.canvas coords $graphEntryId]
    set x [lindex $textXY 0]
    set y [lindex $textXY 1]

    set oldName [.graph.canvas itemcget $graphEntryTextId -text]
    set userName($oldName)          0
    set userName($editNodeUserName) 1

    .graph.canvas itemconfigure $graphEntryTextId -text $editNodeUserName
    .graph.canvas.entryName selection clear
    .graph.canvas coords $graphEntryTextId $x $y
    .graph.canvas coords $graphEntryId -10 -10

    global graphStandAlone editNodeName editNodeType
    if { $graphStandAlone == 0 } {

	if { ($editNodeType == "instance") || ($editNodeType == "group") || \
		($editNodeType == "sphere") || ($editNodeType == "cone") || \
		($editNodeType == "cylinder") || ($editNodeType == "torus") ||  \
		($editNodeType == "cyclide") } {  

	    puts "slideSetParameters $editNodeName $editNodeType 0 guiID $editNodeUserName"
	    slideSetParameters $editNodeName $editNodeType 0 guiID $editNodeUserName
	}
    }
}

proc Graph:EntryShow { textId nodeName } {
    global graphEntryHasFocus editNodeUserName graphEntryId keepSelected graphEntryTextId

    # clicking entry to entry make sure previous entry is hidden
    if { $graphEntryHasFocus == 1 } {

	set error [Graph:EntryProcess]
	if { $error == 1 } {
	    return
	}
    }

    set graphEntryHasFocus 1
    set graphEntryTextId $textId
    set keepSelected 1
    set nodeId [.graph.canvas find withtag $nodeName]

    Graph:ObjectSelected $nodeId

    set textXY [.graph.canvas coords $textId]
    set x [lindex $textXY 0]
    set y [lindex $textXY 1]
    set editNodeUserName [.graph.canvas itemcget $textId -text]

    .graph.canvas coords $textId -10 -10
    .graph.canvas coords $graphEntryId $x $y
    
    focus .graph.canvas.entryName
    .graph.canvas.entryName icursor end
}

proc Graph:LeftRelease { mouseX mouseY } {
    global canvas dragged

    #puts "LeftRelease"

    if { $dragged == 1 } {
	#puts "LeftRelease dragged"
	set dragged 0
    } else {

	#puts "LeftRelease selected"
	
	set x [expr $mouseX + $canvas(xscroll)]
	set y [expr $mouseY + $canvas(yscroll)]

	set nodeId [.graph.canvas find closest $x $y]

	if { -1 == [lsearch [.graph.canvas gettags $nodeId] movable] } {

	    # this object is not movable
	    return
	}

	global entryHideError
	if { $entryHideError == 0 } {

	    Graph:ObjectSelected $nodeId
	}
    }
}

proc Graph:SLIDEObjectSelected { nodeName } {

    set nodeId [.graph.canvas find withtag $nodeName]

    Graph:ObjectSelected $nodeId
}

proc Graph:ObjectSelected { nodeId } {
    global objectSelected selectedObject nodeSLIDEId graphStandAlone viewerSelected

    puts "Graph:ObjectSelected $nodeId with tags: [.graph.canvas gettags $nodeId]"

    # this is commented out because of viewer/graph selections
    # what axes not to be drawn so have to call get parameters
    if { (($objectSelected == 1) && ($nodeId == $selectedObject) && \
	  ($viewerSelected == 0) )} {
	return
    }

    set objectSelected 1
    set selectedObject $nodeId

    set nodeName $nodeSLIDEId($nodeId)
    set textId   [.graph.canvas find withtag text[subst $nodeName]]
    set userName [.graph.canvas itemcget $textId -text]
    set type [lindex [.graph.canvas gettags $nodeId] 0]

    Graph:DrawHighlight $nodeId

    if { $graphStandAlone == 0 } {

	Editor:EditNode $nodeName $userName $type    
    }   
}

proc Graph:ObjectUnselected { callSlide } {
    global graphStandAlone selectedObject objectSelected
    
    # Move the highlight
    .graph.canvas coords highlightNode -5 -5 -5 -5
    .graph.canvas coords highlightInst -5 -5 -5 -5
    set selectedObject none
    set objectSelected 0
    
    if { $graphStandAlone == 0 } {

	if { $callSlide == 1 } {
	    slideUnselect
	}
	global defaultFrame
	raise $defaultFrame
    }
}

# Adding New Nodes
#########################################################################################

#########################################################################################
# Moving Nodes

proc Graph:LeftSelect { mouseX mouseY } {
    global canvas imageWidth imageHeight instanceRadius keepSelected

    # clicking node make sure previous entry is hidden
    global graphEntryHasFocus
    if { $graphEntryHasFocus == 1 } {

	set error [Graph:EntryProcess]
	if { $error == 1 } {
	    return
	}
    }

    set x [expr $mouseX + $canvas(xscroll)]
    set y [expr $mouseY + $canvas(yscroll)]

    set nodeId [.graph.canvas find closest $x $y]

    if { -1 == [lsearch [.graph.canvas gettags $nodeId] movable] } {
	
	# this object is not movable
	return
    }

    set keepSelected 1

    # find the closest object in canvas to $x $y and store these values
    set canvas(obj) $nodeId
    set canvas(x) $x
    set canvas(y) $y
}

proc Graph:LeftDragSelected { mouseX mouseY } {
    global imageWidth imageHeight 
    global canvasTotalWidth canvasTotalHeight canvasWidth canvasHeight
    global canvas nodeInstanceTails nodeInstanceHeads
    global .graph.xscroll .graph.yscroll
    global nodeSLIDEId dragged selectedObject objectSelected

#    puts "LeftDragSelected"
    
    set dragged 1
    set nodeId $canvas(obj)
    set isThisAnInstance [string first instance $[lindex [.graph.canvas gettags $nodeId] 0]]

    if { $isThisAnInstance != -1 } {
	
	#puts "Not moving because this is an instance"
	return
    }
	
    set x [expr $mouseX + $canvas(xscroll)]
    set y [expr $mouseY + $canvas(yscroll)]
    set dx [expr $x - $canvas(x)]
    set dy [expr $y - $canvas(y)]
    
    set nodeXY [.graph.canvas coords $nodeId]
    set xMinCur [lindex $nodeXY 0]
    set yMinCur [lindex $nodeXY 1]
    set xMaxCur [expr $xMinCur + $imageWidth]
    set yMaxCur [expr $yMinCur + $imageHeight]
    
    set xMin [expr $xMinCur + $dx]
    set yMin [expr $yMinCur + $dy]
    set xMax [expr $xMaxCur + $dx]
    set yMax [expr $yMaxCur + $dy]

    ###############################################################    
    # Move the object but make sure its inside the canvas

    # if icon is at a x border don't move it in x unless mouse is in canvas
    if { (($xMinCur == 0) && ($mouseX < 0)) || \
	    (($xMaxCur == $canvasTotalWidth) && ($mouseX > $canvasWidth)) } {

	set dx 0

    } elseif { $xMin < 0 } {

	set dx [expr $dx - $xMin]
	
    } elseif { $xMax > $canvasTotalWidth } {
	
	set dx [expr $dx - $xMax + $canvasTotalWidth]
    }

    # if icon is at a y border don't move it in y unless mouse is in canvas
    if { (($yMinCur == 0) && ($mouseY < 0)) || \
	    (($yMaxCur == $canvasTotalHeight) && ($mouseY > $canvasHeight)) } {
	
	set dy 0

    } elseif { $yMin < 0 } {

	set dy [expr $dy - $yMin]

    } elseif { $yMax > $canvasTotalHeight } {

	set dy [expr $dy - $yMax + $canvasTotalHeight]
    }

    set textName $nodeSLIDEId($nodeId)
    set textId [.graph.canvas find withtag text[subst $textName]]

    .graph.canvas move $nodeId $dx $dy
    .graph.canvas move $textId $dx $dy
    set canvas(x) $x
    set canvas(y) $y

    # redraw all the instances connecting to this node
    set mynodeInstanceTails $nodeInstanceTails($nodeId)
    set mynodeInstanceHeads $nodeInstanceHeads($nodeId)
    
    Graph:RedrawHeadNodeInstances $nodeId 
    Graph:RedrawTailNodeInstances $nodeId 

    # highlight object
    if { $objectSelected == 1 } {
	Graph:DrawHighlight $selectedObject
    }

    ###############################################################
    # automatic scrolling

    set xMin [expr $xMinCur + $dx]
    set yMin [expr $yMinCur + $dy]
    set xMax [expr $xMaxCur + $dx]
    set yMax [expr $yMaxCur + $dy]

    set canvasfXX [.graph.canvas xview]
    set canvasfYY [.graph.canvas yview]    
    set canvasXMin [expr [lindex $canvasfXX 0] * $canvasTotalWidth]
    set canvasXMax [expr [lindex $canvasfXX 1] * $canvasTotalWidth]
    set canvasYMin [expr [lindex $canvasfYY 0] * $canvasTotalHeight]
    set canvasYMax [expr [lindex $canvasfYY 1] * $canvasTotalHeight]
    
    # x-direction
    if { $xMin < $canvasXMin } {

	set first [expr $xMin / $canvasTotalWidth]
	.graph.canvas xview moveto $first	    

    } elseif { $xMax > $canvasXMax } {

	set oldFirst [lindex [.graph.xscroll get] 0]
	set first [expr $oldFirst + ($xMax/$canvasXMax -1)]
	.graph.canvas xview moveto $first	    
    }
    
    # y-direction
    if { $yMin < $canvasYMin } { 
	
	set first [expr $yMin / $canvasTotalHeight]
	.graph.canvas yview moveto $first	    
	
    } elseif { $yMax > $canvasYMax } {
	
	set oldFirst [lindex [.graph.yscroll get] 0]
	set first [expr $oldFirst + ($yMax/$canvasYMax -1)]
	.graph.canvas yview moveto $first	    
    }
}

# Moving Nodes
#########################################################################################

#########################################################################################
# Instancing

proc Graph:CanvasMiddleSelect { mouseX mouseY } {
    global canvas tailNodeId 

    #puts "Graph:CanvasMiddleSelect"

    set x [expr $mouseX + $canvas(xscroll)]
    set y [expr $mouseY + $canvas(yscroll)]
    set tailNodeId [.graph.canvas find closest $x $y]
    
    if { (-1 == [lsearch [.graph.canvas gettags $tailNodeId] group]) && \
	    (-1 == [lsearch [.graph.canvas gettags $tailNodeId] render]) } {

	# this object is not a group or render node
	return
    }

    Graph:CreateInstanceLine $tailNodeId $x $y
}

proc Graph:CreateInstanceLine { tailNodeId x y } {
    global instanceName imageWidth imageHeight instanceTailHeight arrowColor
    global xTail yTail

    set tailXY [.graph.canvas coords $tailNodeId]
    set xTail [expr [lindex $tailXY 0] + $imageWidth/2.0]
    set yTail [expr [lindex $tailXY 1] + $instanceTailHeight]
    set xHead $x
    set yHead $y
    set xMid [expr ($xTail+$xHead)/2.0]
    set yMid [expr ($yTail+$yHead)/2.0]

    set tags instanceLine
    set instanceName instance[subst [Graph:GetNodeId "instance"]]
    set instanceLineName [subst $instanceName]Line
    lappend tags $instanceLineName

    # Line
    .graph.canvas create line $xTail $yTail $xMid $yMid $xHead $yHead \
	    -fill $arrowColor \
	    -joinstyle round -smooth true -width 1 -arrow last -arrowshape {15 15 4} \
	    -tag $tags

#    set instanceId [.graph.canvas find withtag $instanceName]
#    puts "created line $instanceId"
}

proc Graph:CanvasMiddleDrag { mouseX mouseY } {
    global canvas instanceName xTail yTail
    global canvasTotalWidth canvasTotalHeight canvasWidth canvasHeight

    if { $instanceName == "none" } {

	# instance was not started
	return
    }

    set x [expr $mouseX + $canvas(xscroll)]
    set y [expr $mouseY + $canvas(yscroll)]

    # Draw line while mouse is moving
    set xHead $x
    set yHead $y
    set xMid [expr ($xTail+$xHead)/2.0]
    set yMid [expr ($yTail+$yHead)/2.0]

    .graph.canvas coords [subst $instanceName]Line $xTail $yTail \
			    $xMid $yMid \
			    $xHead $yHead 

    ###############################################################
    # automatic scrolling

    set canvasfXX [.graph.canvas xview]
    set canvasfYY [.graph.canvas yview]    
    set canvasXMin [expr [lindex $canvasfXX 0] * $canvasTotalWidth]
    set canvasXMax [expr [lindex $canvasfXX 1] * $canvasTotalWidth]
    set canvasYMin [expr [lindex $canvasfYY 0] * $canvasTotalHeight]
    set canvasYMax [expr [lindex $canvasfYY 1] * $canvasTotalHeight]

    # x-direction
    if { $xHead < $canvasXMin } {
	
	set first [expr $xHead / $canvasTotalWidth]
	.graph.canvas xview moveto $first	    

    } elseif { $xHead > $canvasXMax } {

	set oldFirst [lindex [.graph.xscroll get] 0]
	set first [expr $oldFirst + ($xHead/$canvasXMax -1)]
	.graph.canvas xview moveto $first	    
    }
    
    # y-direction
    if { $yHead < $canvasYMin } { 
	
	set first [expr $yHead / $canvasTotalHeight]
	.graph.canvas yview moveto $first	    
	
    } elseif { $yHead > $canvasYMax } {
	
	set oldFirst [lindex [.graph.yscroll get] 0]
	set first [expr $oldFirst + ($yHead/$canvasYMax -1)]
	.graph.canvas yview moveto $first	    
    }
}

proc Graph:CanvasMiddleRelease { mouseX mouseY } {
    global canvas nodeSLIDEId instanceName tailNodeId xTail yTail 

    if { $instanceName == "none" } {

	# instance was not started
	return
    }

    set x [expr $mouseX + $canvas(xscroll)]
    set y [expr $mouseY + $canvas(yscroll)]

    set instanceId [.graph.canvas find withtag [subst $instanceName]Line]

    # Get node below drawn instance
    set headNodeId [.graph.canvas find closest $x $y 0 $instanceId]

    if { $headNodeId == $instanceId } {

	puts "deleting $instanceId"
	.graph.canvas delete [subst $instanceName]Line
	set instanceName "none"
	return
    }

    set tailName $nodeSLIDEId($tailNodeId)
    set headName $nodeSLIDEId($headNodeId)

    # Only instance correct nodes
    if { ($tailName != $headName) && \
	    (-1 != [lsearch [.graph.canvas gettags $tailNodeId] group]) && \
	    (-1 != [lsearch [.graph.canvas gettags $headNodeId] instancable]) } {
	
	# Create Instance Circle
	Graph:CreateInstanceCircle $instanceName $tailNodeId $headNodeId
	
    } elseif { -1 != [lsearch [.graph.canvas gettags $tailNodeId] render] } {
	
	# Create Instance Circle and Fix head of arrow
	.graph.canvas delete [subst $instanceName]Line

	Graph:CreateReference $tailName $headName
	Graph:RedrawHeadNodeInstances $headNodeId
    } else {
	
	puts "deleting $instanceId"
	.graph.canvas delete [subst $instanceName]Line
	set instanceName "none"
	return
    }
    
    global graphStandAlone
    if { $graphStandAlone == 0 } {
	set headNodeName [lindex [.graph.canvas gettags $headNodeId] 1]
	set tailNodeName [lindex [.graph.canvas gettags $tailNodeId] 1]


	#puts "----------------------- slideAddInstance $instanceName $tailNodeName $headNodeName"

	#slideAddInstance $instanceName $tailNodeName $headNodeName
        slide create instance $instanceName $tailNodeName $headNodeName
    }

    set instanceName "none"
}

proc Graph:RegisterInstance { inst tailNodeId headNodeId } {
    global nodeInstanceTails nodeInstanceHeads instances instanceHead instanceTail

    lappend nodeInstanceTails($tailNodeId) $inst
    lappend nodeInstanceHeads($headNodeId) $inst
    lappend instances($tailNodeId,$headNodeId) $inst
    set instanceHead($inst) $headNodeId
    set instanceTail($inst) $tailNodeId
}

proc Graph:CreateReference { tailName headName } {
    global arrowColor

    set tailNodeId [.graph.canvas find withtag $tailName]
    set headNodeId [.graph.canvas find withtag $headName]

    set tags reference
    set referenceName reference[subst $tailName][subst $headName]
    lappend tags $referenceName

    # Line
    .graph.canvas create line -10 -10 -10 -10 -10 -10 -fill $arrowColor \
	    -joinstyle round -smooth true -width 3 -arrow last -arrowshape {15 15 4} \
	    -tag $tags \

    # RegisterInstance
    Graph:RegisterInstance $referenceName $tailNodeId $headNodeId
    Graph:RedrawHeadNodeInstances $headNodeId
}

proc Graph:CreateInstanceCircle { instName tailNodeId headNodeId } {
    global imageWidth instances instanceTail instanceWidth instanceRadius slideName
    global nodeSLIDEId textHighlightColor userName

    Graph:RegisterInstance $instName $tailNodeId $headNodeId

    set tags instance
    lappend tags $instName
    lappend tags movable

    # Circle
    set instCircleId [.graph.canvas create oval -10 -10 -10 -10 -fill gray -width 0 \
	    -tag $tags]
    set nodeSLIDEId($instCircleId) $instName
    set userName($instName) 1

    # Text
    set tags text[subst $instName]
    set textId [.graph.canvas create text -10 -10 -text $instName -justify center \
	    -tag $tags]
    .graph.canvas bind $textId <Button-1>   "Graph:EntryShow $textId $instName"

    Graph:RedrawHeadNodeInstances $headNodeId
}

proc Graph:CreateInstance { instName tailName headName } {
    global arrowColor

    set tailNodeId [.graph.canvas find withtag $tailName]
    set headNodeId [.graph.canvas find withtag $headName]

    set tags instanceLine
    lappend tags [subst $instName]Line

    # Line
    .graph.canvas create line -10 -10 -10 -10 -10 -10 -fill $arrowColor \
	    -joinstyle round -smooth true -width 1 -arrow last -arrowshape {15 15 4} \
	    -tag $tags \
	
    # InstanceCircle
    Graph:CreateInstanceCircle $instName $tailNodeId $headNodeId
}

# Instancing
#########################################################################################

#########################################################################################
# Update Drawing 

proc Graph:DrawHighlight { nodeId } {
    global imageWidth imageHeight instanceRadius

    # highlight object
    set isThisAnInstance [string first instance $[lindex [.graph.canvas gettags $nodeId] 0]]
    set nodeXY [.graph.canvas coords $nodeId]

    if { $isThisAnInstance != -1 } {
	
	set ovalXMin [lindex $nodeXY 0]
	set ovalYMin [lindex $nodeXY 1]
	set ovalXMax [expr $ovalXMin + 2*$instanceRadius]
	set ovalYMax [expr $ovalYMin + 2*$instanceRadius]
	
	.graph.canvas coords highlightInst $ovalXMin $ovalYMin $ovalXMax $ovalYMax 
	.graph.canvas coords highlightNode -5 -5 -5 -5

    } else {
	
	set rectXMin [lindex $nodeXY 0]
	set rectYMin [lindex $nodeXY 1]
	set rectXMax [expr $rectXMin + $imageWidth]
	set rectYMax [expr $rectYMin + $imageHeight]
	
	.graph.canvas coords highlightNode $rectXMin $rectYMin $rectXMax $rectYMax 
	.graph.canvas coords highlightInst -5 -5 -5 -5
    }
}

proc Graph:RedrawHeadNodeInstances { headNodeId } {
    global nodeInstanceHeads 

    set mynodeInstanceHeads $nodeInstanceHeads($headNodeId)

    foreach inst $mynodeInstanceHeads {
	if { $inst != "empty" } {

	    set isReference [string first reference $inst]
	 
	    if { $isReference == -1 } {
		Graph:DrawInstance $inst
	    } else {
		Graph:DrawReference $inst
	    }
	}
    }
}

proc Graph:RedrawTailNodeInstances { tailNodeId } {
    global nodeInstanceTails 

    set mynodeInstanceTails $nodeInstanceTails($tailNodeId)

    foreach inst $mynodeInstanceTails {
	if { $inst != "empty" } {

	    set isReference [string first reference $inst]

	    if { $isReference == -1 } {
		Graph:DrawInstance $inst
	    } else {
		Graph:DrawReference $inst
	    }
	}
    }
}

proc Graph:DrawInstance { instName } {
    global instances instanceTail instanceHead
    global imageWidth instanceWidth instanceRadius instanceHeadHeight instanceTailHeight
    global selectedObject

    set tailNodeId $instanceTail($instName)
    set headNodeId $instanceHead($instName)

    set tailXY [.graph.canvas coords $tailNodeId]
    set headXY [.graph.canvas coords $headNodeId]
    set xTail [expr [lindex $tailXY 0] + $imageWidth/2.0]
    set yTail [expr [lindex $tailXY 1] + $instanceTailHeight]
    set xHead [expr [lindex $headXY 0] + $imageWidth/2.0]
    set yHead [expr [lindex $headXY 1] + $instanceHeadHeight]

    set xMid [expr ($xTail + $xHead)/2]
    set yMid [expr ($yTail + $yHead)/2]

    set instNum [lsearch $instances($tailNodeId,$headNodeId) $instName]
    set width [expr pow(-1,$instNum) * int(($instNum+1)/2) * $instanceWidth]
    set xMid [expr $xMid + $width]

    # Line
    .graph.canvas coords [subst $instName]Line $xTail $yTail $xMid $yMid $xHead $yHead 

    # Circle
#    set instanceCircleName [subst $instName]Circle
    set instanceCircleName $instName
    set offset -$width/2.0
    set x1 [expr $xMid - $instanceRadius + $offset]
    set y1 [expr $yMid - $instanceRadius]
    set x2 [expr $xMid + $instanceRadius + $offset]
    set y2 [expr $yMid + $instanceRadius]
    .graph.canvas coords $instanceCircleName $x1 $y1 $x2 $y2

    # Highlight
#    if { $selectedObject == [.graph.canvas find withtag $instName] } {
#	Graph:DrawHighlight $selectedObject
#    }

    # Text
    set textName text[subst $instName]
    set oldTextXY [.graph.canvas coords $textName]
    set xOld [lindex $oldTextXY 0]
    set yOld [lindex $oldTextXY 1]

    set xNew [expr $xMid + $offset]
    set pow [expr ceil($instNum/2.0)]
    set yNew [expr $yMid - pow(-1,$pow)*($instanceRadius+10)]

    set dx [expr $xNew - $xOld]
    set dy [expr $yNew - $yOld]

    .graph.canvas move $textName $dx $dy
}

proc Graph:DrawReference { refName } {
    global instanceTail instanceHead
    global imageWidth instanceWidth instanceHeadHeight instanceTailHeight

    set tailNodeId $instanceTail($refName)
    set headNodeId $instanceHead($refName)

    set tailXY [.graph.canvas coords $tailNodeId]
    set headXY [.graph.canvas coords $headNodeId]
    set xTail [expr [lindex $tailXY 0] + $imageWidth/2.0]
    set yTail [expr [lindex $tailXY 1] + $instanceTailHeight]
    set xHead [expr [lindex $headXY 0] + $imageWidth/2.0]
    set yHead [expr [lindex $headXY 1] + $instanceHeadHeight]
    
    set xMid [expr ($xTail + $xHead)/2.0]
    set yMid [expr ($yTail + $yHead)/2.0]

    # Line
    .graph.canvas coords $refName $xTail $yTail $xMid $yMid $xHead $yHead 
}

# Update Drawing
#########################################################################################

#########################################################################################
# Scrolling
#
# convention:
#
# mouseX, mouseY: always (0,0) to (width, height)
#
# x, y: because scroll bars change coordinates need keep track of items by absolute
# coords (0,0) to (totalwidth, totalheight)
#

proc Graph:ScrollX { c fx1 fx2 } {
    global canvas paletteSelection canvasTotalWidth

    # update the scrollbar
    $c.xscroll set $fx1 $fx2

    # remember xposition of canvas
    set x [expr $fx1*$canvasTotalWidth]

    set canvas(xscroll) $x
#    puts "xscroll $x   fx1=$fx1 fx2=$fx2 w=$canvasTotalWidth"
}

proc Graph:ScrollY { c fy1 fy2 } {
    global canvas paletteSelection canvasTotalHeight

    # update the scrollbar
    $c.yscroll set $fy1 $fy2

    # remember yposition of canvas
    set y [expr $fy1*$canvasTotalHeight]

    set canvas(yscroll) $y
#    puts "yscroll $y   fy1=$fy1 fy2=$fy2 h=$canvasTotalHeight"
}

# Scrolling
#########################################################################################

#########################################################################################
# Frames

proc Graph:CreateScrolledCanvas { base width height region } {
    global paletteSelection canvasWidth canvasHeight canvasTotalHeight canvasTotalWidth
    global progress progressText
    
    set paletteSelection none
    set canvasWidth $width
    set canvasHeight $height
    set canvasTotalWidth [lindex $region 2]
    set canvasTotalHeight [lindex $region 3]
    
    incr progress
    set progressText "Creating canvas ..."

    ###################################################################
    # canvas    

    canvas $base.canvas -width $width -height $height -scrollregion $region \
	    -borderwidth 2 -relief ridge \
	    -xscrollcommand [list Graph:ScrollX $base] \
	    -yscrollcommand [list Graph:ScrollY $base]

    ###################################################################
    # scroll bars

    #                                       this moves the canvas objects
    scrollbar $base.yscroll -orient vertical -command [list $base.canvas yview]    

    # Create extra frame to hold pad and horizontal scrollbar
    frame .graph.bottom
    scrollbar $base.xscroll -orient horizontal -command [list $base.canvas xview]
    
    # Create padding based on the scrollbar width and border
    set pad [expr [$base.yscroll cget -width] + 2*([$base.yscroll cget -bd] + \
	    [$base.yscroll cget -highlightthickness])]
    frame $base.pad -width $pad -height $pad

    incr progress
    set progressText "Binding events ..."
    
    ######################################################################
    # Bind actions to objects
    
    # moving nodes
    .graph.canvas bind movable <Button-1>        {Graph:LeftSelect %x %y}
    .graph.canvas bind movable <B1-Motion>       {Graph:LeftDragSelected %x %y}
    .graph.canvas bind movable <ButtonRelease-1> {Graph:LeftRelease %x %y}

    # placing new node
    # selecting multiple nodes
    bind .graph.canvas <Button-1> {Graph:LeftDown %x %y}
    bind .graph.canvas <Motion>   {Graph:Motion %x %y}
        
    # instancing
    .graph.canvas bind movable <Button-2>        {Graph:CanvasMiddleSelect %x %y}
    .graph.canvas bind movable <B2-Motion>       {Graph:CanvasMiddleDrag %x %y}
    .graph.canvas bind movable <ButtonRelease-2> {Graph:CanvasMiddleRelease %x %y}

    # keys
    bind .graph <Key-Delete>    {Graph:Delete %x %y}
    bind .graph <Key-BackSpace> {Graph:Delete %x %y}

    #bind all <Any-Key> {puts %b}
    #bind all <Key-Up> {puts "pressed up"}
    #bind all <Key-Down> {puts "pressed down"}
    #bind all <Key-Left> {puts "pressed left"}
    #bind all <Key-Right> {puts "pressed right"}
    #bind .graph.canvas <Enter> {puts "Entered %W at %x %y"}
    #bind .graph.canvas <Leave> {puts "Left %W at %x %y"}
    
    incr progress
    set progressText "Creating buttons ..."

    ###################################################################
    # buttons
    frame .graph.top -borderwidth 2 -relief raised

    Graph:CreateButtons .graph.top

    incr progress
    set progressText "Packing frames ..."

    ###################################################################
    # Arrange
    pack $base.top -side top -fill x
    pack $base.bottom -side bottom -fill x
    pack $base.pad -in $base.bottom -side right
    pack $base.xscroll -in $base.bottom -side bottom -fill x
    pack $base.yscroll -side right -fill y
    pack $base.canvas -side left -fill both -expand true
#    pack $base -side top -fill both -expand true
    
    incr progress
    set progressText "Graph finished ..."

    return $base.canvas
}

# Frames
#########################################################################################

proc Graph:Init { } {

    Graph:InitCanvasSetting
    Graph:InitImages

    global selectedObject selectedPath objectSelected keepSelected dragged graphStandAlone
    global instanceName graphEntryHasFocus userName entryHideCalled entryHideError
    global viewerSelected

    set selectedObject  none
    set selectedPath    {}
    set objectSelected  0
    set keepSelected    0
    set dragged         0
    set graphStandAlone 0
    set instanceName    none
    set viewerSelected  0

    set graphEntryHasFocus 0
    set userName(1)        1
    set entryHideCalled    0
    set entryHideError     0
}

proc Graph:Create { } {
    global canvasFrame textHighlightColor progress progressText guiSystem

    set w .graph
    toplevel $w
    wm title    $w "Scene Graph"

    if { $guiSystem == "irix" } {
	wm geometry $w +548+67
    } elseif { $guiSystem == "nt" } {
	wm geometry $w +550+65
    } else {
	puts "ERROR Editor:Create: System not IRIX or NT"
	exit
    }	


    wm minsize  $w  515 200
    wm protocol $w WM_DELETE_WINDOW {set foo 1}

    incr progress
    set progressText "Init Graph..."

    Graph:Init

    incr progress
    set progressText "Creating canvas..."

    set canvasFrame [Graph:CreateScrolledCanvas $w 600 785 {0 0 1000 1000}]
    $canvasFrame create rectangle -5 -5 -5 -5 -width 5 -outline red -tag highlightNode
    $canvasFrame create arc -5 -5 -5 -5 -style arc -start 0 \
	    -extent -359 -width 8 -outline red -tag highlightInst

    global editNodeUserName graphEntryId
    Entry .graph.canvas.entryName -textvariable editNodeUserName \
	    -relief flat -width 0 -selectbackground $textHighlightColor \
	    -command "Graph:EntryProcess"

    set graphEntryId [.graph.canvas create window -10 -10 -tag "entryName" \
	    -window .graph.canvas.entryName]

    incr progress
    set progressText "done..."

#    AddDefaultNodes
}

puts "loaded C:/Slide/lib/gui/graph/graph.tcl"